#include "xplode.h"
#include "xplodeshared.h"
#include "xplodeicons.h"
#include "xplodeexport.h"
#include "xplodesearch.h"
#include "xplodereplace.h"
#include "xplodehelp.h"
#include "xplodelogo.h"

#include <qprogressdialog.h>

extern QProgressDialog *initDialog;

// Constructor
XplodeApp::XplodeApp(QString path)
{
        // Set caption and icon
        setCaption("Xplode 2.2");
        setIcon(QPixmap(qembed_findImage("xplode_icon")));

        // Set defaults
        untitledCount = 0;
        pDocList = new QList<XplodeDoc>();
        pDocList->setAutoDelete(true);

        // Initialize the application
        initDefaults(path);
        initView();
        initActions();
        initMenuBar();
        initToolBar();
        initStatusBar();
        initDialog->setTotalSteps(100);
        initDialog->setProgress(100);

        // Initialize the search and replace archive (empty)
        oldTargets = new QStringList();
        oldReplaces = new QStringList();

        // Open a connection with Plasm interpreter
        if (listener->openConnection())
        {
                slotStartEvaluation();
                statusText->setText("Starting Plasm interpreter...");
        }
}

// Destructor
XplodeApp::~XplodeApp()
{
}

// Set the defaults reading the configuration file
void XplodeApp::initDefaults(QString path)
{
        // Get the configuration
        config = new XplodeConfig(path);
        // If an error occurs reading the configuration file, rebuild a new one
        if (!config->readConfig())
        {
                // If an error occurs generating the configuration file
                if (!config->rebuildConfig())
                {
                        QMessageBox::critical(this, "Xplode - Initialization",
                                                "Error creating the configuration file.\nXplode will now quit.");
                        slotFileQuit();
                }
        }
}

// Initialize actions connected to menu and toolbar items
void XplodeApp::initActions()
{
        // Initialize actions and connect them. Self explicative... ;)
        fileNew = new QAction("New File", QPixmap(qembed_findImage("filenew")), "&New", QKeySequence("Ctrl+N"), this);
        fileNew->setStatusTip("Creates a new document");
        fileNew->setWhatsThis("<p><b>New File</b></p> Creates a new Plasm program");
        connect(fileNew, SIGNAL(activated()), this, SLOT(slotFileNew()));

        fileOpen = new QAction("Open File", QPixmap(qembed_findImage("fileopen")), "&Open...", QKeySequence("Ctrl+O"), this);
        fileOpen->setStatusTip("Opens an existing document");
        fileOpen->setWhatsThis("<p><b>Open File</b></p> Opens an existing Plasm program");
        connect(fileOpen, SIGNAL(activated()), this, SLOT(slotFileOpen()));

        fileSave = new QAction("Save File", QPixmap(qembed_findImage("filesave")), "&Save", QKeySequence("Ctrl+S"), this);
        fileSave->setStatusTip("Saves the current Plasm program");
        fileSave->setWhatsThis("<p><b>Save File</b></p> Saves the current Plasm program");
        connect(fileSave, SIGNAL(activated()), this, SLOT(slotFileSave()));

        fileSaveAs = new QAction("Save File As", "Save &As...", 0, this);
        fileSaveAs->setStatusTip("Saves the current Plasm program with a new name");
        fileSaveAs->setWhatsThis("<p><b>Save As</b></p> Saves the current Plasm program with a new file name");
        connect(fileSaveAs, SIGNAL(activated()), this, SLOT(slotFileSaveAs()));

        fileClose = new QAction("Close File", QPixmap(qembed_findImage("fileclose")), "&Close", QKeySequence("Ctrl+W"), this);
        fileClose->setStatusTip("Closes the current Plasm program");
        fileClose->setWhatsThis("<p><b>Close File</b></p> Closes the current Plasm program");
        connect(fileClose, SIGNAL(activated()), this, SLOT(slotFileClose()));

        fileQuit = new QAction("Quit", QPixmap(qembed_findImage("filequit")), "&Quit...", QKeySequence("Ctrl+Q"), this);
        fileQuit->setStatusTip("Quits Xplode");
        fileQuit->setWhatsThis("<p><b>Quit</b></p> Quits the Xplode application");
        connect(fileQuit, SIGNAL(activated()), this, SLOT(slotFileQuit()));

        editCut = new QAction("Cut", QPixmap(qembed_findImage("editcut")), "Cu&t", QKeySequence("Ctrl+X"), this);
        editCut->setStatusTip("Cuts the selected text");
        editCut->setWhatsThis("<p><b>Cut</b></p> Cuts the selected text and puts it into the clipboard");
        connect(editCut, SIGNAL(activated()), this, SLOT(slotEditCut()));
        editCut->setEnabled(false);

        editCopy = new QAction("Copy", QPixmap(qembed_findImage("editcopy")), "&Copy", QKeySequence("Ctrl+C"), this);
        editCopy->setStatusTip("Copies the selected text");
        editCopy->setWhatsThis("<p><b>Copy</b></p> Copies the selected text into the clipboard");
        connect(editCopy, SIGNAL(activated()), this, SLOT(slotEditCopy()));
        editCopy->setEnabled(false);

        editUndo = new QAction("Undo", QPixmap(qembed_findImage("editundo")), "&Undo", QKeySequence("Ctrl+Z"), this);
        editUndo->setStatusTip("Reverts the last action");
        editUndo->setWhatsThis("<p><b>Undo</b></p> Reverts the last editing action");
        connect(editUndo, SIGNAL(activated()), this, SLOT(slotEditUndo()));
        editUndo->setEnabled(false);

        editPaste = new QAction("Paste", QPixmap(qembed_findImage("editpaste")), "&Paste", QKeySequence("Ctrl+V"), this);
        editPaste->setStatusTip("Pastes the clipboard contents");
        editPaste->setWhatsThis("<p><b>Paste</b></p> Pastes the clipboard contents to the current position");
        connect(editPaste, SIGNAL(activated()), this, SLOT(slotEditPaste()));

        editSelectAll = new QAction("SelectAll", "Select &All", QKeySequence("Ctrl+A"), this);
        editSelectAll->setStatusTip("Selects all the text");
        editSelectAll->setWhatsThis("<p><b>Select All</b></p> Selects all the text in the current Plasm program");
        connect(editSelectAll, SIGNAL(activated()), this, SLOT(slotEditSelectAll()));

        editSearch = new QAction("Search", QPixmap(qembed_findImage("editfind")), "&Find...", QKeySequence("Ctrl+F"), this);
        editSearch->setStatusTip("Searches a specified text");
        editSearch->setWhatsThis("<p><b>Find</b></p>Searches a specified text in the current Plasm program");
        connect(editSearch, SIGNAL(activated()), this, SLOT(slotEditSearch()));

#ifdef __APPLE__
        editSearchNext = new QAction("Search Next", QPixmap(qembed_findImage("editfindnext")), "Find &Next...", QKeySequence("Ctrl+G"), this);
#else
        editSearchNext = new QAction("Search Next", QPixmap(qembed_findImage("editfindnext")), "Find &Next...", QKeySequence("F3"), this);
#endif
        editSearchNext->setStatusTip("Searches the next occurrency");
        editSearchNext->setWhatsThis("<p><b>Find Next</b></p> Searches the next occurrency of a specified text");
        connect(editSearchNext, SIGNAL(activated()), this, SLOT(slotEditSearchNext()));

        editReplace = new QAction("Replace", "&Replace...", QKeySequence("Ctrl+R"), this);
        editReplace->setStatusTip("Replaces a specific text");
        editReplace->setWhatsThis("<p><b>Replace</b></p> Replaces a specified text in the current Plasm program");
        connect(editReplace, SIGNAL(activated()), this, SLOT(slotEditReplace()));

        prefPlasm = new QAction("Local", QPixmap(qembed_findImage("plasminterpreter")), "Plasm &Interpreter...", 0, this);
        prefPlasm->setStatusTip("Selects the Plasm interpreter");
        prefPlasm->setWhatsThis("<p><b>Plasm Interpreter</b></p> Sets the Plasm interpreter to be used in the next session");
        connect(prefPlasm, SIGNAL(activated()), this, SLOT(slotPrefPlasm()));

        prefVrml = new QAction("Vrml Viewer", QPixmap(qembed_findImage("prefvrml")), "&Vrml Viewer...", 0, this);
        prefVrml->setStatusTip("Selects the Vrml viewer");
        prefVrml->setWhatsThis("<p><b>Vrml Viewer</b></p> Sets the Vrml viewer to be used");
        connect(prefVrml, SIGNAL(activated()), this, SLOT(slotPrefVrml()));

        prefSvg = new QAction("Svg Viewer", QPixmap(qembed_findImage("prefsvg")), "&Svg Viewer...", 0, this);
        prefSvg->setStatusTip("Selects the Svg viewer");
        prefSvg->setWhatsThis("<p><b>Svg Viewer</b></p> Sets the Svg viewer to be used");
        connect(prefSvg, SIGNAL(activated()), this, SLOT(slotPrefSvg()));

        prefEditFont = new QAction("Editor Font", QPixmap(qembed_findImage("prefeditfont")), "&Editor Font...", 0, this);
        prefEditFont->setStatusTip("Changes the editor font");
        prefEditFont->setWhatsThis("<p><b>Editor Font</b></p> Changes the editor font");
        connect(prefEditFont, SIGNAL(activated()), this, SLOT(slotPrefEditFont()));

        prefListenerFont = new QAction("Listener Font", QPixmap(qembed_findImage("preflistenfont")), "&Listener Font...", 0, this);
        prefListenerFont->setStatusTip("Changes the listener font");
        prefListenerFont->setWhatsThis("<p><b>Listener Font</b></p> Changes the Plasm listener font");
        connect(prefListenerFont, SIGNAL(activated()), this, SLOT(slotPrefListenerFont()));

        prefSave = new QAction("Save Settings", QPixmap(qembed_findImage("filesave")), "&Save Settings", 0, this);
        prefSave->setStatusTip("Saves the current settings");
        prefSave->setWhatsThis("<p><b>Save Settings</b></p> Saves current Xplode settings");
        connect(prefSave, SIGNAL(activated()), this, SLOT(slotPrefSave()));

        plasmExport = new QAction("Export", QPixmap(qembed_findImage("plasmexport")), "E&xport...", 0, this);
        plasmExport->setStatusTip("Exports a polyhedral complex");
        plasmExport->setWhatsThis("<p><b>Export</b></p> Exports a polyhedral complex into various file formats");
        connect(plasmExport, SIGNAL(activated()), this, SLOT(slotPlasmExport()));

        plasmVrml = new QAction("Vrml Export", QPixmap(qembed_findImage("plasmvrml")), "&VRML Export...", QKeySequence("Ctrl+M"), this);
        plasmVrml->setStatusTip("Exports a polyhedral complex into a VRML file");
        plasmVrml->setWhatsThis("<p><b>Vrml Export</b></p> Exports a polyhedral complex into a VRML file");
        connect(plasmVrml, SIGNAL(activated()), this, SLOT(slotPlasmVrml()));

        plasmConnect = new QAction("Start Plasm interpreter", QPixmap(qembed_findImage("plasmconnect")), "&Start Plasm...", 0, this);
        plasmConnect->setStatusTip("Starts the Plasm interpreter");
        plasmConnect->setWhatsThis("<p><b>Start Plasm</b></p> Starts a new Plasm interpreter session");
        connect(plasmConnect, SIGNAL(activated()), this, SLOT(slotPlasmConnect()));

        plasmExpr = new QAction("Evaluate Expression", QPixmap(qembed_findImage("plasmevalexpr")), "Evaluate &Expression", QKeySequence("Ctrl+E"), this);
        plasmExpr->setStatusTip("Evaluates a selected expression");
        plasmExpr->setWhatsThis("<p><b>Evaluate Expression</b></p> Evaluates a selected expression");
        connect(plasmExpr, SIGNAL(activated()), this, SLOT(slotPlasmExpr()));
        plasmExpr->setEnabled(false);

        plasmBuffer = new QAction("Evaluate Buffer", QPixmap(qembed_findImage("plasmevalbuffer")), "Evaluate &Buffer", QKeySequence("Ctrl+B"), this);
        plasmBuffer->setStatusTip("Evaluates the entire editor buffer");
        plasmBuffer->setWhatsThis("<p><b>Evaluate Buffer</b></p> Evaluates the entire editor buffer");
        connect(plasmBuffer, SIGNAL(activated()), this, SLOT(slotPlasmBuffer()));
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If there is one, check
        if (editor)
        {
                // Get the text
                QString code = editor->text();
                // If there is no text, disable the action
                if (code.isEmpty()) plasmBuffer->setEnabled(false);
        }

        helpHelp = new QAction("Help", QPixmap(qembed_findImage("helphelp")), "&Help...", QKeySequence("F1"), this);
        helpHelp->setStatusTip("Opens the Xplode help");
        helpHelp->setWhatsThis("<p><b>Help</b></p> Opens the Xplode help system");
        connect(helpHelp, SIGNAL(activated()), this, SLOT(slotHelpHelp()));

        helpAboutApp = new QAction("About", QPixmap(qembed_findImage("helpabout")), "&About...", 0, this);
        helpAboutApp->setStatusTip("Shows the Xplode credits");
        helpAboutApp->setWhatsThis("<p><b>About</b></p> Shows the Xplode credits");
        connect(helpAboutApp, SIGNAL(activated()), this, SLOT(slotHelpAbout()));

        helpAboutQt = new QAction("AboutQt", QPixmap(qembed_findImage("about_qt")), "&About Qt...", 0, this);
        helpAboutQt->setStatusTip("Shows the Qt Library credits");
        helpAboutQt->setWhatsThis("<p><b>About Qt</b></p>Shows the Qt Library credits");
        connect(helpAboutQt, SIGNAL(activated()), this, SLOT(slotHelpAboutQtLib()));
}

// Initialize the menu bar
void XplodeApp::initMenuBar()
{
        // Menu 'File'
        pFileMenu = new QPopupMenu();
        fileNew->addTo(pFileMenu);
        fileOpen->addTo(pFileMenu);
        fileClose->addTo(pFileMenu);
        pFileMenu->insertSeparator();
        fileSave->addTo(pFileMenu);
        fileSaveAs->addTo(pFileMenu);
        pFileMenu->insertSeparator();
        fileQuit->addTo(pFileMenu);

        // Menu 'Edit'
        pEditMenu = new QPopupMenu();
        editUndo->addTo(pEditMenu);
        pEditMenu->insertSeparator();
        editCut->addTo(pEditMenu);
        editCopy->addTo(pEditMenu);
        editPaste->addTo(pEditMenu);
        pEditMenu->insertSeparator();
        editSelectAll->addTo(pEditMenu);
        pEditMenu->insertSeparator();
        editSearch->addTo(pEditMenu);
        editSearchNext->addTo(pEditMenu);
        editReplace->addTo(pEditMenu);

        // Menu 'Preferences'
        pPrefMenu = new QPopupMenu();
        prefPlasm->addTo(pPrefMenu);
        // Submenu 'Viewers'
        pPrefViewersMenu = new QPopupMenu();
        prefVrml->addTo(pPrefViewersMenu);
        prefSvg->addTo(pPrefViewersMenu);
        pPrefMenu->insertItem("&Viewers", pPrefViewersMenu);
        // Submenu 'Fonts'
        pPrefFontMenu = new QPopupMenu();
        prefEditFont->addTo(pPrefFontMenu);
        prefListenerFont->addTo(pPrefFontMenu);
        pPrefMenu->insertItem("&Fonts", pPrefFontMenu);
        pPrefMenu->insertSeparator();
        prefSave->addTo(pPrefMenu);

        // Menu 'Plasm'
        pPlasmMenu = new QPopupMenu();
        pPlasmMenu->insertSeparator();
        plasmExport->addTo(pPlasmMenu);
        plasmVrml->addTo(pPlasmMenu);
        pPlasmMenu->insertSeparator();
        plasmConnect->addTo(pPlasmMenu);
        pPlasmMenu->insertSeparator();
        plasmExpr->addTo(pPlasmMenu);
        plasmBuffer->addTo(pPlasmMenu);

        // Menu 'Help'
        pHelpMenu = new QPopupMenu();
        helpHelp->addTo(pHelpMenu);
        pHelpMenu->insertItem(QIconSet(QPixmap(qembed_findImage("whatsthis"))), "&What's This?", this, SLOT(whatsThis()), SHIFT+Key_F1);
        pHelpMenu->insertSeparator();
        helpAboutApp->addTo(pHelpMenu);
        helpAboutQt->addTo(pHelpMenu);

        menuBar()->insertItem("&File", pFileMenu);
        menuBar()->insertItem("&Edit", pEditMenu);
        menuBar()->insertItem("P&references", pPrefMenu);
        menuBar()->insertItem("&Plasm", pPlasmMenu);
        menuBar()->insertItem("&Help", pHelpMenu);
}

// Initialize the toolbar
void XplodeApp::initToolBar()
{
        xplodeToolbar = new QToolBar(this, "xplode operations");
        fileNew->addTo(xplodeToolbar);
        fileOpen->addTo(xplodeToolbar);
        fileSave->addTo(xplodeToolbar);
        xplodeToolbar->addSeparator();
        editCut->addTo(xplodeToolbar);
        editCopy->addTo(xplodeToolbar);
        editPaste->addTo(xplodeToolbar);
        xplodeToolbar->addSeparator();
        editUndo->addTo(xplodeToolbar);
        xplodeToolbar->addSeparator();
        editSearch->addTo(xplodeToolbar);
        editSearchNext->addTo(xplodeToolbar);
        xplodeToolbar->addSeparator();
        plasmConnect->addTo(xplodeToolbar);
        xplodeToolbar->addSeparator();
        plasmExport->addTo(xplodeToolbar);
        plasmVrml->addTo(xplodeToolbar);
        xplodeToolbar->addSeparator();
        plasmExpr->addTo(xplodeToolbar);
        plasmBuffer->addTo(xplodeToolbar);
}

// Initialize the status bar
void XplodeApp::initStatusBar()
{
        // Initialize status bar objects
        statusText = new QLabel(this);
        cursorPos = new QLabel(this);
        abortEval = new QPushButton(QIconSet(QPixmap(qembed_findImage("button_cancel"))), "Abort", this);
        // The button is enabled when evaluating
        abortEval->setEnabled(false);
        evalProgress = new QProgressBar(10, this);
        evalProgress->setPercentageVisible(false);
        evalTimer = new QTimer(this);
        // Add widgets to the status bar: a label, the "Abort" button and the progress bar
        statusBar()->addWidget(statusText, 4);
        statusBar()->addWidget(cursorPos);
        statusBar()->addWidget(abortEval);
        statusBar()->addWidget(evalProgress);
        // Connect signals
        connect(abortEval, SIGNAL(clicked()), this, SLOT(slotKillEvaluation()));
        connect(evalTimer, SIGNAL(timeout()), this, SLOT(slotIncrementEvaluation()));
}

// Initialize the main window view: a splitted window
void XplodeApp::initView()
{
        // Maximized
        setWindowState(WindowMaximized);
        pView = new QSplitter(QSplitter::Vertical, this);
        pWorkspace = new QSplitter(QSplitter::Horizontal, pView);
        pTabbedWorkspace = new QTabWidget(pWorkspace);
        pDefs = new QToolBox(pWorkspace);
        // Add user-definied definitions
        pDefinitions = new XplodeDefinitions(pDefs, "def");
        pDefs->addItem(pDefinitions, "User definitions");
        pDock = new QSplitter(QSplitter::Horizontal, pView);
        // Add the standard plasm definitions
        pFun = new PlasmFunctionListBox(pDefs, "fun");
        pDefs->addItem(pFun, "Plasm Functions");
        // Set the central widget
        setCentralWidget(pView);
        // Create the listener
        listener = new XplodeListener(pDock, "lst");
        // Create the exported objects list box
        pExp = new QToolBox(pDock);
        pExported = new XplodeDefinitions(pExp, "exp");
        pExp->addItem(pExported, "Exported objects");
        // Set all the widget sizes
        int theSize;
        QValueList<int> sizes = pView->sizes();
        theSize = sizes[0] + sizes[1];
        sizes[0] = (int) (80 * theSize / 100);
        sizes[1] = (int) (20 * theSize / 100);
        pView->setSizes(sizes);
        sizes = pWorkspace->sizes();
        theSize = sizes[0] + sizes[1];
        sizes[0] = (int) (80 * theSize / 100);
        sizes[1] = (int) (20 * theSize / 100);
        pWorkspace->setSizes(sizes);
        pDock->setSizes(sizes);
        // Connect signals
        connect(pTabbedWorkspace, SIGNAL(currentChanged(QWidget*)), this, SLOT(slotUpdateMenus()));
        connect(listener, SIGNAL(startEvaluation()), this, SLOT(slotStartEvaluation()));
        connect(listener, SIGNAL(stopEvaluation()), this, SLOT(slotStopEvaluation()));
        connect(listener, SIGNAL(exported(QString)), this, SLOT(slotAddExported(QString)));
        connect(listener, SIGNAL(defined(QString)), this, SLOT(slotAddDefinition(QString)));
        connect(pExported, SIGNAL(doubleClicked(QListBoxItem*)), this, SLOT(slotViewExported(QListBoxItem*)));
        connect(pExported, SIGNAL(dropped(QString)), this, SLOT(slotDropExport(QString)));
        connect(pDefinitions, SIGNAL(doubleClicked(QListBoxItem*)), this, SLOT(slotViewDefinition(QListBoxItem*)));
        connect(pDefinitions, SIGNAL(dropped(QString)), this, SLOT(slotDropDefinition(QString)));
}

// Open a document
void XplodeApp::openDocumentFile(const char* file)
{
        statusText->setText("Opening file...");
        XplodeDoc* doc;
        // Check, if document already open. If so, set the focus to the correct view
        for (doc = pDocList->first(); doc > 0; doc = pDocList->next())
                if (doc->pathName() == file)
                {
                        // The document is already open: inform the user
                        QMessageBox::information(this, "Xplode - Open", "The document is already open.", QMessageBox::Ok | QMessageBox::Default);
                        return;
                }
        // Create an empty document
        doc = new XplodeDoc();
        pDocList->append(doc);
        doc->newDocument();
        // Create an empty editor
        XplodeEditor *editor = new XplodeEditor(pTabbedWorkspace, 0);
        // Connect it to the document
        doc->connectEditor(editor);
        // Creates an untitled document if file is null
        if (!file)
        {
                untitledCount++;
                QString fileName = QString("Untitled%1").arg(untitledCount);
                doc->setPathName(fileName);
                doc->setTitle(fileName);
        }
        // Open the file
        else
                // If an error occurs opening the file
                if (!doc->openDocument(file))
                {
                        QMessageBox::critical(this, "Xplode - Error", "An error occurred opening the selected file.",  QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
                        pDocList->remove(doc); // was: delete doc;
                        delete editor;
                        return;
                }
        // Connect signals
        connect(doc, SIGNAL(documentModified(XplodeDoc*)), this, SLOT(slotFileModified(XplodeDoc*)));
        connect(doc, SIGNAL(documentSaved(XplodeDoc*)), this, SLOT(slotFileSaved(XplodeDoc*)));
        connect(doc, SIGNAL(documentCursor(QString)), this, SLOT(slotShowCursorPosition(QString)));
        connect(editor, SIGNAL(selectionChanged()), this, SLOT(slotUpdateMenus()));
        connect(editor, SIGNAL(undoAvailable(bool)), this, SLOT(slotUpdateMenus()));
        // Show the document in the tabbed view
        pTabbedWorkspace->addTab(editor, doc->title());
        // Show the editor and set the focus
        pTabbedWorkspace->showPage(editor);
        editor->setFocus();
        this->slotShowCursorPosition("L: 1, C: 0");
        statusText->setText("Ready.");
}

// Ask whether to quit
void XplodeApp::closeEvent(QCloseEvent *e)
{
        slotFileQuit();
}

// Return true if the user wants to quit
bool XplodeApp::queryExit()
{
        bool canQuit = true;
        XplodeDoc* doc;
        // Check if a document has been modified
        for (doc = pDocList->first(); doc > 0; doc = pDocList->next())
                if (doc->isModified())
                {
                        canQuit = false;
                        break;
                }
        // If so, ask whether do quit
        if (!canQuit)
                canQuit = QMessageBox::warning(this, "Xplode - Quit", "One or more files have been modified.\nDo your really want to quit?", QMessageBox::Yes, QMessageBox::No | QMessageBox::Default) == QMessageBox::Yes;
        return canQuit;
}
// Create a new document
void XplodeApp::slotFileNew()
{
        statusText->setText("Creating new file...");
        openDocumentFile();
        statusText->setText("Ready.");
}

// Open an existing document
void XplodeApp::slotFileOpen()
{
        statusText->setText("Opening file...");
        // Get the file name
        QString fileName = QFileDialog::getOpenFileName(0, plasmDocuments, this);
        // If the file name is valid, open the document
        if (!fileName.isEmpty()) openDocumentFile(fileName);
        statusText->setText("Ready.");
}

// Save the current document
void XplodeApp::slotFileSave()
{
        statusText->setText("Saving file...");
        // Get the current editor
        XplodeEditor *currentEditor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If null, so no file is open, exit
        if (currentEditor == 0) return;
        // Search for the corresponding document
        XplodeDoc *doc, *currentDoc = 0;
        for (doc = pDocList->first(); doc > 0; doc = pDocList->next())
                if (doc->getEditor() == currentEditor) currentDoc = doc;
        // If no document is found, something has gone *really bad*
        if (!currentDoc)
        {
                QMessageBox::critical(this, "Xplode - Error", "Cannot find a valid file pointer.\nThis is a <b>serious</b> situation.", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
                return;
        }
        // If the document is unnamed, ask for the name
        if (currentDoc->title().contains("Untitled"))
                slotFileSaveAs();
        else
                // If an error occurs saving the document, show a message
                if (!currentDoc->saveDocument(currentDoc->pathName()))
                        QMessageBox::critical(this, "Xplode - Error", "Cannot save the current document.", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
        statusText->setText(tr("Ready."));
}

// Save the current document with a different name
void XplodeApp::slotFileSaveAs()
{
        statusText->setText("Saving file as...");
        // Get the current editor
        XplodeEditor *currentEditor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If null, so no file is open, exit
        if (currentEditor == 0) return;
        // Search for the corresponding document
        XplodeDoc *doc, *currentDoc = 0;
        for (doc = pDocList->first(); doc > 0; doc = pDocList->next())
                if (doc->getEditor() == currentEditor) currentDoc = doc;
        // If no document is found, something has gone *really bad*
        if (!currentDoc)
        {
                QMessageBox::critical(this, "Xplode - Error", "Cannot find a valid document pointer.\nThis is a <b>serious</b> situation.", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
                return;
        }
        // Ask for a name
        QString fileName = QFileDialog::getSaveFileName(currentDoc->pathName(), plasmDocuments, this);
        // If the name is not valid, exit
        if (fileName.isEmpty()) return;
        if(!currentDoc->saveDocument(fileName)) QMessageBox::critical(this, "Xplode - Error", "Cannot save the current document.", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
        // Set tab name
        pTabbedWorkspace->setTabLabel(currentEditor, currentDoc->title());
        statusText->setText(tr("Ready."));
}

// Close the current document
void XplodeApp::slotFileClose()
{
        statusText->setText("Closing file...");
        // Get the current editor
        XplodeEditor *currentEditor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If null, so no file is open, exit
        if (currentEditor == 0) return;
        // Search for the corresponding document
        XplodeDoc *doc, *currentDoc = 0;
        for (doc = pDocList->first(); doc > 0; doc = pDocList->next())
                if (doc->getEditor() == currentEditor) currentDoc = doc;
        // If no document is found, something has gone *really bad*
        if (!currentDoc)
        {
                QMessageBox::critical(this, "Xplode - Error", "Cannot find a valid document pointer.\nThis is a <b>serious</b> situation.", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
                return;
        }
        if (currentDoc->canCloseDocument())
        {
                // Disconnect signals
                connect(currentDoc, SIGNAL(documentModified(XplodeDoc*)), this, SLOT(slotFileModified(XplodeDoc*)));
                connect(currentDoc, SIGNAL(documentSaved(XplodeDoc*)), this, SLOT(slotFileSaved(XplodeDoc*)));
                connect(currentDoc, SIGNAL(documentCursor(QString)), this, SLOT(slotShowCursorPosition(QString)));
                connect(currentEditor, SIGNAL(selectionChanged()), this, SLOT(slotUpdateMenus()));
                connect(currentEditor, SIGNAL(undoAvailable(bool)), this, SLOT(slotUpdateMenus()));
                pDocList->remove(currentDoc);
                pTabbedWorkspace->removePage(currentEditor);
                delete currentEditor;
                statusText->setText("Ready.");
        }
        slotUpdateMenus();
}

// Quit Xplode
void XplodeApp::slotFileQuit()
{
        statusText->setText("Exiting application...");
        // If we must quit, do it
        if (queryExit())
        {
                listener->closeConnection();
                qApp->quit();
        }
        statusText->setText("Ready.");
}

// Undo last action
void XplodeApp::slotEditUndo()
{
        statusText->setText("Reverting last action...");
        // Get the current editor
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If there is one, undo
        if (editor) editor->undo();
        statusText->setText("Ready.");
}

// Cut selection into the clipboard
void XplodeApp::slotEditCut()
{
        statusText->setText("Cutting selection to clipboard...");
        // If the listener has focus
        if (listener->hasFocus())
        {
                // The listener cannot be modified: a cut is actually a copy
                listener->copy();
        }
        else
        {
                // Get the current editor
                XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
                // If there is one, cut
                if (editor) editor->cut();
        }
        statusText->setText("Ready.");
}

// Copy selection into the clipboard
void XplodeApp::slotEditCopy()
{
        statusText->setText("Copying selection to clipboard...");
        // If the listener has focus
        if (listener->hasFocus())
        {
                // The listener cannot be modified: a cut is actually a copy
                listener->copy();
        }
        else
        {
                // Get the current editor
                XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
                // If there is one, copy
                if (editor) editor->copy();
        }
        statusText->setText("Ready.");
}

// Insert clipboard content into the current document
void XplodeApp::slotEditPaste()
{
        statusText->setText("Inserting clipboard content...");
        // Get the current editor
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If there is one, paste
        if (editor) editor->paste();
        statusText->setText("Ready.");
}

// Select all the text in the current editor
void XplodeApp::slotEditSelectAll()
{
        statusText->setText("Selecting all text...");
        // Get the current editor
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        if (!editor) return;
        // If there is one, select all the text
        int lastParag = editor->paragraphs() - 1;
        int lastIndex = editor->text(lastParag).length();
        editor->setSelection(0, 0, lastParag, lastIndex);
        statusText->setText("Ready.");
}

// Search for a specific text
void XplodeApp::slotEditSearch()
{
        statusText->setText("Finding text...");
        // Get the current editor
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        if (!editor) return;
        XplodeSearch *searchDlg = new XplodeSearch(this);
        // Insert search history
        searchDlg->textTarget->insertStringList(*oldTargets);
        searchDlg->textTarget->setCurrentText("");
        searchDlg->setFocusProxy(searchDlg->textTarget);
        // Show the search dialog box
        if (searchDlg->exec())
        {
                // Set global variables
                target = searchDlg->textTarget->currentText();
                replace = "";
                bRegexp = searchDlg->boxRegexp->isChecked();
                bCase = searchDlg->boxCase->isChecked();
                bBack = searchDlg->boxBack->isChecked();
                bCursor = searchDlg->boxCursor->isChecked();
                bAll = false;
                int startIdx, startPar, curIdx, curPar, found;
                QString para, sTarget(target);
                QRegExp rTarget(target);
                rTarget.setCaseSensitive(bCase);
                // Check the validity of the regular expression
                if (bRegexp)
                {
                        if (!rTarget.isValid())
                        {
                                QString msg = "Invalid regular expression:\n\n%1";
                                // Show a message
                                QMessageBox::warning(this, "Xplode - Find", msg.arg(rTarget.errorString()), QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
                                // Invalidate the action
                                target = "";
                                // Exit
                                return;
                        }
                }
                // Check if history has to be updated: a new search?
                if (oldTargets->find(target) == oldTargets->end())
                {
                        // Max 5 items
                        if (oldTargets->count() == 5) oldTargets->remove(oldTargets->last());
                        // Update history
                        oldTargets->prepend(target);
                }
                // Set initial cursor position
                if (bCursor)
                        editor->getCursorPosition(&startPar, &startIdx);
                else
                {
                        startPar = 0;
                        startIdx = 0;
                }
                // Set current cursor position
                curPar = startPar;
                curIdx = startIdx;
                // Set current paragraph string
                para = editor->text(curPar);
                // Set found to start the search
                found = -1;
                // Search
                while (found == -1)
                {
                        // Search the current paragraph
                        if (bRegexp)
                        {
                                if (bBack)
                                        found = rTarget.searchRev(para, curIdx);
                                else
                                        found = rTarget.search(para, curIdx);
                        }
                        else
                        {
                                if (bBack)
                                        found = para.findRev(sTarget, curIdx, bCase);
                                else
                                        found = para.find(sTarget, curIdx, bCase);
                        }
                        // If we found something
                        if (found != -1)
                        {
                                int foundLen;
                                // Get the found substring length
                                if (bRegexp)
                                        foundLen = rTarget.capturedTexts().first().length();
                                else
                                        foundLen = target.length();

                                // Highlight the substring and move cursor
                                if (bBack)
                                {
                                        editor->setSelection(curPar, found, curPar, found + foundLen);
                                        editor->setCursorPos(curPar, found);
                                }
                                else
                                {
                                        editor->setSelection(curPar, found, curPar, found + foundLen);
                                        editor->setCursorPos(curPar, found + foundLen);
                                }
                        }
                        else
                        {
                                // Go to the next paragraph
                                if (bBack)
                                {
                                        curPar -= 1;
                                        // Check paragraph validity
                                        if (curPar == -1) curPar = editor->paragraphs() - 1;
                                        para = editor->text(curPar);
                                        curIdx = para.length();
                                }
                                else
                                {
                                        curPar += 1;
                                        // Check paragraph validity
                                        if (curPar >= editor->paragraphs()) curPar = 0;
                                        para = editor->text(curPar);
                                        curIdx = 0;
                                }
                                // Check the starting paragraph (the last one)
                                if (curPar == startPar)
                                {
                                        // Search the current paragraph
                                        if (bRegexp)
                                        {
                                                if (bBack)
                                                        found = rTarget.searchRev(para, curIdx);
                                                else
                                                        found = rTarget.search(para, curIdx);
                                        }
                                        else
                                        {
                                                if (bBack)
                                                        found = para.findRev(sTarget, curIdx, bCase);
                                                else
                                                        found = para.find(sTarget, curIdx, bCase);
                                        }
                                        // Nothing has been found
                                        if (found == -1)
                                        {
                                                QMessageBox::warning(this, "Xplode - Find", "Search string not found.", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
                                                // Exit the while loop
                                                found = 0;
                                                // Invalidate last search
                                                target = "";
                                        }
                                }
                        }
                }
        }
        statusText->setText("Ready.");
}

// Search for a specific text
void XplodeApp::slotEditSearchNext()
{
        statusText->setText("Finding next occurrency...");
        // Get the current editor
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        if (!editor) return;
        // If no search has been done, ask for a new one
        if (target == "")
                slotEditSearch();
        else
        {
                int startIdx, startPar, curIdx, curPar, found;
                QString para, sTarget(target);
                QRegExp rTarget(target);
                rTarget.setCaseSensitive(bCase);
                // Get initial cursor position
                editor->getCursorPosition(&startPar, &startIdx);
                // Adjust position if finding backwards
                if (bBack)
                {
                        // Adjust index
                        startIdx -= 1;
                        // Check its the validity
                        if (startIdx < 0)
                        {
                                // Adjust paragraph
                                startPar -= 1;
                                // Check its validity
                                if (startPar < 0) startPar = editor->paragraphs() - 1;
                                startIdx = editor->text(startPar).length();
                        }
                }
                // Set current cursor position
                curPar = startPar;
                curIdx = startIdx;
                // Set current paragraph string
                para = editor->text(curPar);
                // Set found to start the search
                found = -1;
                // Search
                while (found == -1)
                {
                        // Search the current paragraph
                        if (bRegexp)
                        {
                                if (bBack)
                                        found = rTarget.searchRev(para, curIdx);
                                else
                                        found = rTarget.search(para, curIdx);
                        }
                        else
                        {
                                if (bBack)
                                        found = para.findRev(sTarget, curIdx, bCase);
                                else
                                        found = para.find(sTarget, curIdx, bCase);
                        }
                        // If we found something
                        if (found != -1)
                        {
                                int foundLen;
                                // Get the found substring length
                                if (bRegexp)
                                        foundLen = rTarget.capturedTexts().first().length();
                                else
                                        foundLen = target.length();

                                // Highlight the substring and move cursor
                                if (bBack)
                                {
                                        editor->setSelection(curPar, found, curPar, found + foundLen);
                                        editor->setCursorPos(curPar, found);
                                }
                                else
                                {
                                        editor->setSelection(curPar, found, curPar, found + foundLen);
                                        editor->setCursorPos(curPar, found + foundLen);
                                }
                        }
                        else
                        {
                                // Go to the next paragraph
                                if (bBack)
                                {
                                        curPar -= 1;
                                        // Check paragraph validity
                                        if (curPar == -1) curPar = editor->paragraphs() - 1;
                                        para = editor->text(curPar);
                                        curIdx = para.length();
                                }
                                else
                                {
                                        curPar += 1;
                                        // Check paragraph validity
                                        if (curPar >= editor->paragraphs()) curPar = 0;
                                        para = editor->text(curPar);
                                        curIdx = 0;
                                }
                        }
                }
        }
        statusText->setText("Ready.");
}

// Search for a specific text
void XplodeApp::slotEditReplace()
{
        statusText->setText("Replacing text...");
        // Get the current editor
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        if (!editor) return;
        XplodeReplace *replaceDlg = new XplodeReplace(this);
        // Insert search history
        replaceDlg->textTarget->insertStringList(*oldTargets);
        replaceDlg->textReplace->insertStringList(*oldReplaces);
        replaceDlg->textTarget->setCurrentText("");
        replaceDlg->textReplace->setCurrentText("");
        replaceDlg->setFocusProxy(replaceDlg->textTarget);
        // User abort
        bool aborted = false;
        // Found and replaced items count
        int  nFound, nReplaced;
        // Show the replace dialog box
        if (replaceDlg->exec())
        {
                // Set global variables
                target = replaceDlg->textTarget->currentText();
                replace = replaceDlg->textReplace->currentText();
                bRegexp = replaceDlg->boxRegexp->isChecked();
                bCase = replaceDlg->boxCase->isChecked();
                bBack = replaceDlg->boxBack->isChecked();
                bCursor = replaceDlg->boxCursor->isChecked();
                bAll = replaceDlg->boxAll->isChecked();
                int startIdx, startPar, curIdx, curPar, found, offset = 0;
                QString para, sTarget(target), invariant = "";
                QRegExp rTarget(target);
                rTarget.setCaseSensitive(bCase);
                bool firstRun = true, mustStop = false;
                // Check the validity of the regular expression
                if (bRegexp)
                {
                        if (!rTarget.isValid())
                        {
                                QString msg = "Invalid regular expression:\n\n%1";
                                // Show a message
                                QMessageBox::warning(this, "Xplode - Replace", msg.arg(rTarget.errorString()), QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
                                // Invalidate the action
                                target = "";
                                // Exit
                                return;
                        }
                }
                // Check if history has to be updated: a new replace?
                if (oldTargets->find(target) == oldTargets->end())
                {
                        // Max 5 items
                        if (oldTargets->count() == 5) oldTargets->remove(oldTargets->last());
                        // Update history
                        oldTargets->prepend(target);
                }
                if (oldReplaces->find(replace) == oldReplaces->end())
                {
                        // Max 5 items
                        if (oldReplaces->count() == 5) oldReplaces->remove(oldReplaces->last());
                        // Update history
                        oldReplaces->prepend(replace);
                }
                // Set initial cursor position
                if (bCursor)
                        editor->getCursorPosition(&startPar, &startIdx);
                else
                {
                        startPar = 0;
                        startIdx = 0;
                }
                // Set the current paragraph
                curPar = startPar;
                // Set current paragraph string
                para = editor->text(curPar);
                para = para.left(para.length() - 1);
                // Adjust paragraph
                if (bBack)
                {
                        // Set the invariant part of the paragraph
                        int len = para.length() - startIdx;
                        if (len > 0)
                                invariant = para.right(len);
                        else
                                invariant = "";
                        // Take the left part
                        para = para.left(startIdx);
                        // Set the starting index
                        curIdx = startIdx;
                        // Set the offset
                        offset = 0;
                }
                else
                {
                        // Set the invariant part of the paragraph
                        if (startIdx > 0)
                                invariant = para.left(startIdx);
                        else
                                invariant = "";
                        // Take the right part
                        para = para.right(para.length() - startIdx);
                        // Set the starting index
                        curIdx = 0;
                        // Set the offset
                        offset = startIdx;
                }
                // Set the found/replaced counter to 0
                nFound = nReplaced = 0;
                // Set found to start the search
                found = -1;
                // Search
                while (found == -1)
                {
                        // Check if we're back on the first paragraph
                        if ((curPar == startPar) && (!firstRun) && (!mustStop))
                        {
                                // Stop on next paragraph change
                                mustStop = true;
                                QString tmp = editor->text(startPar);
                                // Set the correct offset and index
                                if (bBack)
                                {
                                        curIdx = invariant.length();
                                        offset = para.length() - curIdx;
                                        // Update the invariant part
                                        int len = tmp.length() - invariant.length();
                                        if (len > 0)
                                                invariant = tmp.left(len);
                                        else
                                                invariant = "";
                                        // Update the paragraph
                                        para = tmp.mid(tmp.length() - invariant.length() - 1);
                                }
                                else
                                {
                                        curIdx = 0;
                                        offset = 0;
                                        // Update the invariant part
                                        int len = tmp.length() - invariant.length();
                                        if (len > 0)
                                                invariant = tmp.right(len);
                                        else
                                                invariant = "";
                                        // Update the paragraph
                                        para = tmp.mid(0, tmp.length() - invariant.length());
                                }
                                // Now use start coordinates
                                startIdx = curIdx;
                        }
                        // Search the current paragraph
                        if (bRegexp)
                        {
                                if (bBack)
                                        found = rTarget.searchRev(para, curIdx);
                                else
                                        found = rTarget.search(para, curIdx);
                        }
                        else
                        {
                                if (bBack)
                                        found = para.findRev(sTarget, curIdx, bCase);
                                else
                                        found = para.find(sTarget, curIdx, bCase);
                        }
                        // If we found something
                        if (found != -1)
                        {
                                // Increase the number of found items
                                nFound++;
                                int foundLen;
                                // Get the found substring length
                                if (bRegexp)
                                        foundLen = rTarget.capturedTexts().first().length();
                                else
                                        foundLen = target.length();

                                // Highlight the substring and move cursor
                                if (bBack)
                                {
                                        editor->setSelection(curPar, found + offset, curPar, found + foundLen + offset);
                                        editor->setCursorPos(curPar, found + offset);
                                }
                                else
                                {
                                        editor->setSelection(curPar, found + offset, curPar, found + foundLen + offset);
                                        editor->setCursorPos(curPar, found + foundLen + offset);
                                }
                                // Update position and index
                                editor->getCursorPos(&curPar, &curIdx);
                                // Replace without inquiry?
                                if (bAll)
                                {
                                        // Replace the text
                                        editor->insert(replace);
                                        // Update the current paragraph
                                        para = editor->text(curPar);
                                        para = para.left(para.length() - 1);
                                        // Update the current index if going forward
                                        if (!bBack) editor->setCursorPos(curPar, found + replace.length());
                                        editor->getCursorPos(&curPar, &curIdx);
                                        // Take care if we're back on the first paragraph
                                        if (mustStop)
                                        {
                                                if (bBack)
                                                        para = para.mid(para.length() - invariant.length() - 1);
                                                else
                                                        para = para.mid(0, para.length() - invariant.length());
                                        }
                                        // Increase the number of replaced items
                                        nReplaced++;
                                        // Continue searching
                                        found = -1;
                                }
                                else
                                {
                                        QString msg = "Replace this occurrence?";
                                        int ret = QMessageBox::information(this, "Xplode - Replace", msg, QMessageBox::Cancel, QMessageBox::No, QMessageBox::Yes | QMessageBox::Default);
                                        // If the user wants to replace the found item, do it
                                        if (ret == QMessageBox::Yes)
                                        {
                                                // Replace the text
                                                editor->insert(replace);
                                                // Update the current paragraph
                                                para = editor->text(curPar);
                                                para = para.left(para.length() - 1);
                                                // Update the current index if going forward
                                                if (!bBack) editor->setCursorPos(curPar, found + replace.length());
                                                editor->getCursorPos(&curPar, &curIdx);
                                                // Take care if we're back on the first paragraph
                                                if (mustStop)
                                                {
                                                        if (bBack)
                                                                para = para.mid(para.length() - invariant.length() - 1);
                                                        else
                                                                para = para.mid(0, para.length() - invariant.length());
                                                }
                                                // Increase the number of replaced items
                                                nReplaced++;
                                                }
                                        // Continue searching
                                        found = -1;
                                        // If the user aborts the operation
                                        if (ret == QMessageBox::Cancel)
                                        {
                                                // Exit from the loop
                                                found = 0;
                                                // Set aborted operation flag
                                                aborted = true;
                                        }
                                }
                                // Update the current index if going backward
                                if (bBack)
                                {
                                        // Adjust index
                                        curIdx -= 1;
                                        // Check its the validity
                                        if (curIdx < 0)
                                        {
                                                // Adjust paragraph
                                                curPar -= 1;
                                                // Check its validity
                                                if (curPar < 0) curPar = editor->paragraphs() - 1;
                                                para = editor->text(curPar);
                                                para = para.left(para.length() - 1);
                                                curIdx = para.length();
                                                if (editor->paragraphs() == 1) firstRun = false;
                                        }
                                }
                        }
                        else
                        {
                                // Go to the next paragraph
                                if (bBack)
                                {
                                        curPar -= 1;
                                        // Check paragraph validity
                                        if (curPar == -1) curPar = editor->paragraphs() - 1;
                                        para = editor->text(curPar);
                                        para = para.left(para.length() - 1);
                                        curIdx = para.length();
                                }
                                else
                                {
                                        curPar += 1;
                                        // Check paragraph validity
                                        if (curPar >= editor->paragraphs()) curPar = 0;
                                        para = editor->text(curPar);
                                        para = para.left(para.length() - 1);
                                        curIdx = 0;
                                }
                                // The first paragraph has been checked. The next loop over the paragraph is the last action
                                firstRun = false;
                                // The offset is useless now
                                offset = 0;
                                // If we already checked the whole first paragraph we must stop
                                if (mustStop) found = 0;
                        }
                }
                // If the user didn't abort the operation show statistics
                if (!aborted)
                {
                        // If nothing has been found
                        if (nFound == 0)
                                QMessageBox::warning(this, "Xplode - Replace", "Search string not found.", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
                        else
                        {
                                QString msg = "Found %1 entries (%2 replaced).";
                                QMessageBox::information(this, "Xplode - Replace", msg.arg(nFound).arg(nReplaced), QMessageBox::Ok | QMessageBox::Default);
                        }
                }
                // Invalidate last search
                target = "";
        }
        statusText->setText("Ready.");
}

// Plasm interpreter path
void XplodeApp::slotPrefPlasm()
{
        statusText->setText("Plasm location...");
        // Get the Plasm executable
        QString selectedFile = QFileDialog::getOpenFileName(*plasmPath, executableFiles, this, 0, "Xplode - Plasm Interpreter");
        // If the user accepts changes, change default value
        if (!selectedFile.isEmpty()) *plasmPath = selectedFile;
        statusText->setText("Ready.");
}

// Vrml viewer
void XplodeApp::slotPrefVrml()
{
        statusText->setText("VRML Viewer location...");
        // Get the VRML viewer executable
        QString selectedFile = QFileDialog::getOpenFileName(*plasmVrmlViewer, executableFiles, this, 0, "Xplode - VRML Viewer");
        // If the user accepts changes, change default value
        if (!selectedFile.isEmpty()) *plasmVrmlViewer = selectedFile;
        statusText->setText("Ready.");
}

// Svg Viewer
void XplodeApp::slotPrefSvg()
{
        statusText->setText("SVG Viewer location...");
        // Get the SVG viewer executable
        QString selectedFile = QFileDialog::getOpenFileName(*plasmSvgViewer, executableFiles, this, 0, "Xplode - SVG Viewer");
        // If the user accepts changes, change default value
        if (!selectedFile.isEmpty()) *plasmSvgViewer = selectedFile;
        statusText->setText("Ready.");
}

// Editor font selection
void XplodeApp::slotPrefEditFont()
{
        statusText->setText("Changing editor font...");
        QFont selectedFont;
        bool fontAccepted;
        // Get the new font
        selectedFont = QFontDialog::getFont(&fontAccepted, *plasmEditorFont, this);
        // If the user accept changes
        if (fontAccepted)
        {
                // Set the new value
                *plasmEditorFont = selectedFont;
                // Update all documents
                XplodeDoc *doc;
                for (doc = pDocList->first(); doc > 0; doc = pDocList->next())
                        doc->getEditor()->configChanged();
        }
        statusText->setText("Ready.");
}

// Listener font selection
void XplodeApp::slotPrefListenerFont()
{
        statusText->setText("Changing listener font...");
        QFont selectedFont;
        bool fontAccepted;
        // Get the new font
        selectedFont = QFontDialog::getFont(&fontAccepted, *plasmListenerFont, this);
        // If the user accept changes
        if (fontAccepted)
        {
                // Set the new value
                *plasmListenerFont = selectedFont;
                // Update the listener
                listener->updateFont();
        }
        statusText->setText("Ready.");
}

// Save preferences
void XplodeApp::slotPrefSave()
{
        statusText->setText("Saving settings...");

        // pWorkspace pDock
        QValueList<int> sizes = pWorkspace->sizes();
        QValueList<int>::iterator it;
        sizes = pDock->sizes();
        // Save settings, on error alert the user
        if (!config->saveConfig()) QMessageBox::critical(this, "Xplode - Error", "Cannot save preferences file.", QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
        statusText->setText("Ready.");
}

// Export a polyhedral complex
void XplodeApp::slotPlasmExport()
{
        statusText->setText("Exporting...");
        QString defaultVal;
        // Get the default export value: if something is selected, then show selection, else 'last_value'
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If no editor is active, exit
        if (!editor)
        {
                statusText->setText("Ready.");
                return;
        }
        if (editor->hasSelectedText())
                defaultVal = editor->selectedText();
        else
                defaultVal = "last_value";
        // Create the dialog
        XplodeExport *exportDlg = new XplodeExport(this);
        // Set the default export object
        exportDlg->textObject->setText(defaultVal);
        // Execute the dialog
        int ret = exportDlg->exec();
        // Get object and file name
        QString exportObj, exportName, exportDir;
        exportObj = exportDlg->textObject->text();
        exportName = exportDlg->textName->text();
        exportDir = exportDlg->textDirectory->text();
        // Build the target file name
        if (!exportDir.startsWith("<")) exportName.prepend(exportDir);
        // If the user accepts and the given name is not empty
        if (ret && (!exportName.isEmpty()))
        {
                // Plasm code to be sent to the server
                QString code;
                // Export format is determined by file extension
                if (exportDlg->radioSuffix->isChecked())
                {
                        // No known suffix: ask whether to change it into a VRML file
                        if ((!exportName.lower().endsWith(".wrl")) && (!exportName.lower().endsWith(".svg")))
                        {
                                QString msg = "Don't know how to handle '%1'. Valid suffixes are .wrl and .svg.\nDo you want to export the file as a VRML document?";
                                int change = QMessageBox::information(this, "Xplode - Export", msg.arg(exportName), QMessageBox::No, QMessageBox::Yes | QMessageBox::Default);
                                if (change == QMessageBox::Yes)
                                        code = "export:(" + exportObj + "):'" + exportName + "';";
                                else
                                        QMessageBox::information(this, "Xplode - Export", "Didn't know how to handle the given file name.\n"\
                                                        "No object has been exported.", QMessageBox::Ok | QMessageBox::Default);
                        }
                        else
                        {
                                // Check the corresponding radio button
                                if (exportName.lower().endsWith(".wrl"))
                                {
                                        exportDlg->radioVrml->setChecked(true);
                                        exportDlg->radioSvg->setChecked(false);
                                }
                                if (exportName.lower().endsWith(".svg"))
                                {
                                        exportDlg->radioVrml->setChecked(false);
                                        exportDlg->radioSvg->setChecked(true);
                                }
                        }
                }
                if (exportDlg->radioVrml->isChecked())
                {
                        // Build the code
                        code = "export:(" + exportObj + "):'" + exportName + "';";
                }
                if (exportDlg->radioSvg->isChecked())
                {
                        // Get the extra parameter
                        QString extra = QInputDialog::getText("Xplode - SVG Export", "The SVG format needs an extra parameter:");
                        // Build the code
                        code = "svg:(" + exportObj + "):" + extra + ":'" + exportName + "';";
                }
                // Send the code, if any
                if (!code.isEmpty()) listener->eval(code);
        }
        statusText->setText("Ready.");
}

// Export a polyhedral complex in VRML directly
void XplodeApp::slotPlasmVrml()
{
        statusText->setText("VRML Export...");
        QString exportObj, exportName = "", defaultVal;
        bool objAccepted;
        // Get the default export value: if something is selected, then show selection, else 'last_value'
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If no editor is active, exit
        if (!editor)
        {
                statusText->setText("Ready.");
                return;
        }
        if (editor->hasSelectedText())
                defaultVal = editor->selectedText();
        else
                defaultVal = "last_value";
        // Get the object to be exported
        exportObj = QInputDialog::getText("Xplode - VRML Export", "Insert the object to export:", QLineEdit::Normal, defaultVal, &objAccepted);
        // If cancel (or ESC) is pressed, exit
        if (!objAccepted) return;
        // If locally connected, ask about the local file name
        exportName = QFileDialog::getSaveFileName(0, vrmlDocuments, this, 0, "VRML Export");
        if (exportName.isEmpty())
        {
                statusText->setText("Ready.");
                return;
        }
        // If no suffix is given add the default vrml extension
        if (!exportName.lower().endsWith(".wrl")) exportName.append(".wrl");
        // Build the export statement
        QString code = "export:(" + exportObj + "):'";
        code += exportName + "';";
        listener->eval(code);
        statusText->setText("Ready.");
}

// Connect to Plasm interpreter
void XplodeApp::slotPlasmConnect()
{
        statusText->setText("Starting Plasm interpreter...");
        int ret = QMessageBox::Yes;
        // If we're connected, ask the user
        if (listener->isConnected())
        {
                ret = QMessageBox::warning(this, "Xplode - Start Plasm", "You are about to start a new Plasm interpreter.\n" \
                                                 "Continuing you'll lose all processed data.\nDo you want to continue?",
                                                 QMessageBox::Yes,
                                                 QMessageBox::No | QMessageBox::Default,
                                                 QMessageBox::NoButton);
        }
        // If the user wants to reconnect, do it
        if (ret == QMessageBox::Yes)
        {
                // Clear the exported objects list
                pExported->clear();
                pDefinitions->clear();
                if (listener->reconnect())
                        // Start the animation
                        slotStartEvaluation();
        }
}

// Evaluate a selected expression
void XplodeApp::slotPlasmExpr()
{
        statusText->setText("Evaluating selected expression...");
        // Get the current editor
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If there is one, continue evaluating
        if (editor)
        {
                // Get the selected text
                QString code = editor->selectedText();
                // If there is a selection, evaluate it
                if (!code.isEmpty())
                {
                        listener->eval(code);
                        slotStartEvaluation();
                }
        }
        statusText->setText("Ready.");
        // Set the focus on the editor
        setFocusProxy(editor);
}

// Evaluate the entire buffer
void XplodeApp::slotPlasmBuffer()
{
        statusText->setText("Evaluating current buffer...");
        // Get the current editor
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If there is one, continue evaluating
        if (editor)
        {
                // Get the text
                QString code = editor->text();
                // If there is a selection, evaluate it
                if (!code.isEmpty())
                {
                        listener->eval(code);
                        slotStartEvaluation();
                }
        }
        statusText->setText("Ready.");
        // Set the focus on the editor
        setFocusProxy(editor);
}

// Help
void XplodeApp::slotHelpHelp()
{
        statusText->setText("Showing Xplode help...");
        XplodeHelp *help = new XplodeHelp(this);
        // Set the source document
        QDir d(*plasmXplode);
        QString absPath = d.absPath();
        // Get the directory
#ifdef WIN32
        QString path = absPath.left(absPath.length() - 10);
#else
        QString path = absPath.left(absPath.length() - 6);
#endif
        help->browserView->setSource(path + "help/index.html");
        // Set the icons
        help->buttonBack->setPixmap(qembed_findImage("helpback"));
        help->buttonForward->setPixmap(qembed_findImage("helpforward"));
        help->buttonHome->setPixmap(qembed_findImage("helphome"));
        help->buttonClose->setPixmap(qembed_findImage("button_cancel"));
        // Show the help window
        help->show();
        statusText->setText("Ready.");
}

// About Xplode
void XplodeApp::slotHelpAbout()
{
        statusText->setText("About Xplode...");
        XplodeLogo *about = new XplodeLogo(this);
        about->show();
        statusText->setText("Ready.");
}

// About TrollTech's Qt Library
void XplodeApp::slotHelpAboutQtLib()
{
        statusText->setText("About the Qt Library...");
        QMessageBox::aboutQt(this, "Xplode - About Qt");
        statusText->setText("Ready.");
}

// Start the animation during evaluations
void XplodeApp::slotStartEvaluation()
{
        // Set the initial progress status
        evalProgress->setProgress(0, 10);
        // Start the timer (animation)
        evalTimer->start(100);
        // The button is enabled when evaluating
        abortEval->setEnabled(true);
}

// Stop the animation
void XplodeApp::slotStopEvaluation()
{
        // Set the initial progress status
        evalProgress->setProgress(0, 10);
        // Stop the timer (animation)
        evalTimer->stop();
        // The button is enabled when evaluating
        abortEval->setEnabled(false);
        statusText->setText("Ready.");
}

// Animate the progress bar
void XplodeApp::slotIncrementEvaluation()
{
        // If we reached the maximum, then decrease the counter
        if (evalProgress->progress() == 10) fwdProgress = false;
        // If we reached the minimum, then increase the counter
        if (evalProgress->progress() == 0) fwdProgress = true;
        // Do the right action
        if (fwdProgress)
                evalProgress->setProgress(evalProgress->progress() + 1);
        else
                evalProgress->setProgress(evalProgress->progress() - 1);

}

// Abort the evaluation: that means killing plasm
void XplodeApp::slotKillEvaluation()
{
        int ret;
        // Ask the user about aborting
        ret = QMessageBox::information(this, "Xplode - Abort", "Plasm must be killed to abort operations." \
                                "Do you really want to continue?",
                                QMessageBox::Yes, QMessageBox::No | QMessageBox::Default);
        // If the user wants to abort
        if (ret == QMessageBox::Yes) listener->closeConnection();
}

void XplodeApp::slotAddExported(QString filename)
{
        // Check if there's already the exported item
        QListBoxItem *item = pExported->findItem(filename);
        if (item)
        {
                // Remove the old item
                pExported->removeItem(pExported->index(item));
        }
        // Add the exported file name to the list, prepending an icon: this is a *new* exported file
        pExported->insertItem(QPixmap(qembed_findImage("newexported")), filename);
        // Set the current item on the last exported file
        pExported->setCurrentItem(pExported->count() - 1);
        // Center the view on the last exported object
        pExported->centerCurrentItem();
}

void XplodeApp::slotAddDefinition(QString defname)
{
        // Check if there's already the exported item
        QListBoxItem *item = pDefinitions->findItem(defname);
        if (item)
        {
                // Remove the old item
                pDefinitions->removeItem(pDefinitions->index(item));
        }
        // Add the new definition
        pDefinitions->insertItem(defname);
        // Center the view on the last definition
        pDefinitions->setCurrentItem(pDefinitions->count() - 1);

}

void XplodeApp::slotViewExported(QListBoxItem* item)
{
        statusText->setText("Viewing file...");
        QString dataType, viewer, fileName = item->text();
        // Set default exporting to VRML
        dataType = "VRML Document";
        // Set the correct VRML viewer
        viewer = "";
        // VRML file
        if (fileName.lower().endsWith(".wrl"))
        {
                dataType = "VRML Document";
                viewer = *plasmVrmlViewer;
        }
        // SVG file
        if (fileName.lower().endsWith(".svg"))
        {
                dataType = "SVG Document";
                viewer = *plasmSvgViewer;
        }
        // No valid suffix has been found
        if (viewer == "")
        {
                QString msg = "Unknown file suffix for '%1'.\nDo you want to view the file as a VRML document?";
                int ret = QMessageBox::information(this, "Xplode - View", msg.arg(fileName), QMessageBox::No, QMessageBox::Yes | QMessageBox::Default);
                // If the user doesn't want to view the unknown file as a VRML, exit
                if (ret == QMessageBox::No) return;
                viewer = *plasmVrmlViewer;
        }
        // Create the viewer process
        QProcess *viewerProcess = new QProcess(this);
#ifdef __APPLE__
        // Apple specific command line
        viewerProcess->addArgument("/usr/bin/open");
        viewerProcess->addArgument("-a");
#endif
        viewerProcess->addArgument(viewer);
        // View the saved file
        viewerProcess->addArgument(fileName);
        // Start the process
        if (!viewerProcess->start())
        {
                // Alert the user about the error
                QMessageBox::warning(this, "Xplode - Error", "Error starting " + dataType + " Viewer:\n" + viewer,
                                                        QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
        }
        // The file has been viewed: remove the new file icon
        pExported->changeItem(fileName, pExported->currentItem());
        statusText->setText("Ready.");
}

void XplodeApp::slotViewDefinition(QListBoxItem* item)
{
        statusText->setText("Evaluating definition...");
        QString defName = item->text();
        slotStartEvaluation();
        listener->eval(defName);
        statusText->setText("Ready.");
}

void XplodeApp::slotFileModified(XplodeDoc* doc)
{
        // Add to the tab an icon
        pTabbedWorkspace->changeTab(doc->getEditor(), QIconSet(QPixmap(qembed_findImage("filemodified"))), doc->title());
        slotUpdateMenus();
}

void XplodeApp::slotFileSaved(XplodeDoc* doc)
{
        // Remove the icon from the tab
        pTabbedWorkspace->changeTab(doc->getEditor(), QIconSet(), doc->title());
}

void XplodeApp::slotShowCursorPosition(QString pos)
{
        cursorPos->setText(pos);
}

void XplodeApp::slotUpdateMenus()
{
        // Get the current editor
        XplodeEditor *editor = (XplodeEditor*) pTabbedWorkspace->currentPage();
        // If there is one, check
        if (editor)
        {
                // Get the text
                QString code = editor->text();
                // If there is no text, disable the action
                if (code.isEmpty())
                        plasmBuffer->setEnabled(false);
                else
                        plasmBuffer->setEnabled(true);
                // If there's a selected text, enable the actions
                if (editor->hasSelectedText())
                {
                        plasmExpr->setEnabled(true);
                        editCopy->setEnabled(true);
                        editCut->setEnabled(true);
                }
                else
                {
                        plasmExpr->setEnabled(false);
                        editCopy->setEnabled(false);
                        editCut->setEnabled(false);
                }
                // If undo items are available, then enable the action
                if (editor->isUndoAvailable())
                        editUndo->setEnabled(true);
                else
                        editUndo->setEnabled(false);
                // Some other actions are enabled by default with a non null editor
                fileSave->setEnabled(true);
                fileSaveAs->setEnabled(true);
                fileClose->setEnabled(true);
                editPaste->setEnabled(true);
                editReplace->setEnabled(true);
                editSearch->setEnabled(true);
                editSearchNext->setEnabled(true);
                editSelectAll->setEnabled(true);
                // Set the focus on the editor
                setFocusProxy(editor);
        }
        else
        {
                // When no editor is available, disable actions
                plasmBuffer->setEnabled(false);
                editUndo->setEnabled(false);
                plasmExpr->setEnabled(false);
                editCopy->setEnabled(false);
                editCut->setEnabled(false);
                editPaste->setEnabled(false);
                fileSave->setEnabled(false);
                fileSaveAs->setEnabled(false);
                editReplace->setEnabled(false);
                editSearch->setEnabled(false);
                editSearchNext->setEnabled(false);
                editSelectAll->setEnabled(false);
                fileClose->setEnabled(false);
        }
}

void XplodeApp::slotDropDefinition(QString text)
{
        statusText->setText("Evaluating selected expression...");
        // If there is a selection, evaluate it
        if (!text.isEmpty())
        {
                listener->eval(text);
                slotStartEvaluation();
        }
        statusText->setText("Ready.");
}

void XplodeApp::slotDropExport(QString text)
{
        statusText->setText("VRML Export...");
        QString exportObj, exportName = "", defaultVal = text;
        bool objAccepted;
        // If locally connected, ask about the local file name
        exportName = QFileDialog::getSaveFileName(0, vrmlDocuments, this, 0, "VRML Export (" + text + ")");
        if (exportName.isEmpty())
        {
                statusText->setText("Ready.");
                return;
        }
        // If no suffix is given add the default vrml extension
        if (!exportName.lower().endsWith(".wrl")) exportName.append(".wrl");
        // Build the export statement
        QString code = "export:(" + defaultVal + "):'";
        code += exportName + "';";
        listener->eval(code);
        statusText->setText("Ready.");
}
